# -*- coding: utf-8 -*-
from setproctitle import setproctitle

from live_client.utils import logging
from live_client.query import on_event
from live_client.events import messenger


__all__ = ["start"]

read_timeout = 120


def start(settings, **kwargs):
    logging.info("Trade frequency monitor started")
    setproctitle("DDA: Trade frequency monitor")

    monitor_settings = settings.get("monitor", {})
    window_duration = monitor_settings.get("window_duration", 60)
    sampling_frequency = monitor_settings.get("sampling_frequency", 30)
    max_threshold = monitor_settings.get("max_threshold", 100)

    fr_query = f"""krakenfx
        => count() by pair every {sampling_frequency} seconds over last {window_duration} seconds
        => @filter(count > {max_threshold})
    """
    span = f"last {window_duration} seconds"

    @on_event(fr_query, settings, span=span, timeout=read_timeout)
    def handle_events(event):
        # Generate alerts whether the threshold was reached
        # a new event means another threshold breach
        event_content = event.get("data", {}).get("content", [])

        for item in event_content:
            template = "{} traded {} times over the last {} seconds"
            message = template.format(item["pair"], int(item["count"]), window_duration)
            messenger.send_message(message, timestamp=item["timestamp"], settings=settings)

        return

    handle_events()



# -*- coding: utf-8 -*-
from setproctitle import setproctitle
import asyncio
import queue
import json
from multiprocessing import Process, Queue

import websockets
from eliot import start_action

from live_client.events import raw
from live_client.utils import logging

__all__ = ["start"]


async def read_results(url, pairs, output_queue):
    setproctitle("krakenfx: reading updates")
    with start_action(action_type="krakenfx.fetch_updates", url=url):
        # connect to the server
        async with websockets.connect(url) as websocket:
            subscription = {"event": "subscribe", "subscription": {"name": "trade"}, "pair": pairs}
            logging.info(f"Subscribing to '{subscription}'")
            await websocket.send(json.dumps(subscription))

            # listen for incoming messages
            async for message in websocket:
                logging.debug(f"New message'{message}'")
                output_queue.put(message)


def watch(url, pairs, output_queue):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(read_results(url, pairs, output_queue))


def get_trades(url, pairs):
    """
    We use a new process to consume the kraken api
    The goal is to isolate the `asyncio` code from the rest of the program
    """
    results_queue = Queue()
    process = Process(target=watch, args=(url, pairs, results_queue))
    process.start()
    return process, results_queue


def start(settings, **kwargs):
    """
    Monitors trades of a set of (crypto)currency pairs using `kraken.com` public api.

    For each trade detected, a new event is sent to live.
    """
    setproctitle("DDA: Currency trades datasource")

    # Input settings
    krakenfx_url = settings.get("krakenfx_url", "wss://ws.kraken.com/")
    pairs = settings.get("pairs", ["ETH/USD", "XBT/USD"])
    timeout = settings.get("timeout", 30)

    # Output settings
    event_type = settings.get("output", {}).get("event_type", "dda_crypto_trades")
    skipstorage = settings.get("output", {}).get("skipstorage", True)

    state_manager = kwargs.get("state_manager")
    state = state_manager.load()
    last_trades = state.get("last_trades", {})

    # A separate process is used to fetch data from kraken
    kraken_process, results_queue = get_trades(krakenfx_url, pairs)

    # Handle the events received from kraken
    while True:
        try:
            trade_data = json.loads(results_queue.get(timeout=timeout))
        except queue.Empty:
            logging.exception(f"No results after {timeout} seconds")
            break

        # We are only interested in trade events
        is_trade = isinstance(trade_data, list) and len(trade_data) == 4
        if is_trade:
            # Prepare an event
            channel_id, operation_data, operation_type, pair = trade_data
            operations = [
                {
                    "price": item[0],
                    "volume": item[1],
                    "time": item[2],
                    "side": item[3],
                    "orderType": item[4],
                    "misc": item[5],
                }
                for item in operation_data
            ]

            trade_event = {
                "channel_id": channel_id,
                "operations": operations,
                "operation_type": operation_type,
                "pair": pair,
                "__skipstorage": skipstorage,
            }

            # And send to live
            raw.create(event_type, trade_event, settings)

            # Update this datasource's state with the last trade for each pair
            # This might be useful if you needed to restore this state
            # when the datasource is restarted
            last_trades.update(pair=trade_event)
            state_manager.save({"last_trades": last_trades})

        else:
            logging.debug(f"Ignoring event {trade_data}")
            continue

    # Release resources on exit
    results_queue.close()
    kraken_process.terminate()
    kraken_process.join()
    return



from chatterbot.conversation import Statement


class ActionStatement(Statement):
    def __init__(self, text, confidence=None, in_response_to=None, **kwargs):
        super().__init__(text, in_response_to, **kwargs)
        self.confidence = confidence
        self._chatbot = kwargs.get("chatbot")

    def run(self):
        raise NotImplementedError()

    def _instance_text(self):
        return f"{self.__class__.__name__}: {id(self)}"

    @property
    def chatbot(self):
        return self._chatbot

    @chatbot.setter
    def chatbot(self, value):
        self.set_chatbot(value)

    def set_chatbot(self, value):
        self._chatbot = value
        return self

    @property
    def live_client(self):
        return self._chatbot.live_client


class ShowTextAction(ActionStatement):
    def run(self):
        return self.text


class NoTextAction(ActionStatement):
    def __init__(self, confidence=None, in_response_to=None, **kwargs):
        super().__init__(self._instance_text(), confidence, in_response_to, **kwargs)
        self.params = kwargs


class CallbackAction(ActionStatement):
    def __init__(self, callback, confidence=None, in_response_to=None, **kwargs):
        super().__init__(self._instance_text(), confidence, in_response_to, **kwargs)
        self.params = kwargs
        self.callback = callback

    def run(self):
        return self.callback(**self.params)


class ChainedAction(ActionStatement):
    def __init__(self, actions, confidence=None, in_response_to=None, **kwargs):
        super().__init__(self._instance_text(), confidence, in_response_to, **kwargs)
        self.actions = actions

    def run(self):
        for action in self.actions:
            try:
                action.run()
            except Exception as e:
                return str(e)



# -*- coding: utf-8 -*-
import time

from live_client.utils import logging

__all__ = ["await_next_cycle"]


def await_next_cycle(sleep_time, message=None, log_func=None):
    if message is None:
        message = "Sleeping for {} seconds".format(sleep_time)

    if log_func is None:
        log_func = logging.debug

    log_func(message)
    time.sleep(sleep_time)



# -*- coding: utf-8 -*-
import time
from hashlib import md5
from typing import Mapping, Dict, Union, AnyStr, Any

import dill
from live_client.utils import logging

__all__ = ["StateManager"]

number = Union[int, float]

TIMESTAMP_KEY = "__timestamp"


class StateManager(object):
    def __init__(self, name: AnyStr, delay_between_updates: number = 60):
        self.name = name
        self.delay_between_updates = delay_between_updates
        self.updated_at = 0

        if isinstance(name, str):
            name = bytes(name, "utf-8")

        self.identifier = md5(name).hexdigest()
        self.filename = f"/tmp/{self.identifier}.{name.decode('utf-8')}.live_agent"

    def load(self) -> Dict[str, Any]:
        state_filename = self.filename

        try:
            with open(state_filename, r"r+b") as f:
                state = dill.load(f)
        except Exception:
            state = {}

        self.updated_at = state.get(TIMESTAMP_KEY, self.updated_at)

        logging.info(f"State for {self.identifier} ({len(state)} keys) loaded")
        return state

    def save(self, state: Mapping[str, Any], force: bool = False) -> None:
        now = time.time()
        next_possible_update = self.updated_at + self.delay_between_updates
        time_until_update = next_possible_update - now

        if (time_until_update > 0) and (not force):
            logging.debug(f"Update for {self.identifier} dropped. Wait {time_until_update:.2f}s")
        else:
            self.do_save(state, timestamp=now)

        return

    def do_save(self, state: Mapping[str, Any], timestamp: number) -> None:
        state_filename = self.filename
        state.update(TIMESTAMP_KEY=timestamp)

        with open(state_filename, r"w+b") as f:
            dill.dump(state, f)

        self.updated_at = timestamp
        logging.debug(f"State for {self.identifier} saved")



# -*- coding: utf-8 -*-
import importlib

from live_client.utils import logging

__all__ = ["load_enabled_modules"]


def log_and_import(name, package=None):
    try:
        return importlib.import_module(name)
    except Exception as e:
        logging.info(f"Error importing {name} (from package={package}): {e}")


def load_enabled_modules(settings):
    modules = []
    for name in settings.get("enabled_modules", []):
        module = log_and_import(name)
        if module is not None:
            modules.append(module)

    return modules


def load_process_handlers(settings):
    enabled_modules = load_enabled_modules(settings)
    process_handlers = {}
    for module in enabled_modules:
        process_handlers.update(**module.PROCESSES)

    return process_handlers



from eliot import start_action  # NOQA
from multiprocessing import active_children
from chatterbot.conversation import Statement

from live_client.utils import logging

from live_agent.services.processes import agent_function
from ..src.actions import CallbackAction
from .base import BaseBayesAdapter, WithAssetAdapter

__all__ = ["MonitorControlAdapter"]

"""
TODO:

1. definir formato das configurações
2. implementar controle para parada dos monitores
3. implementar detecção de dados insuficientes
4. implementar controles individuais para os monitores
"""


class MonitorControlAdapter(BaseBayesAdapter, WithAssetAdapter):
    """
    Controls the monitors for an assets
    """

    state_key = "monitor-control"
    required_state = ["assetId"]
    default_state = {"active_monitors": {}}
    positive_examples = ["start", "run", "monitor", "monitors"]
    description = "Start the monitors"
    usage_example = "start the monitors"

    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)
        self.room_id = kwargs.get("room_id")
        self.settings = kwargs.get("settings", {})
        self.process_handlers = self.settings.get("process_handlers")
        self.all_monitors = self.settings.get("monitors", {})

    def process(self, statement, additional_response_selection_parameters=None):
        confidence = self.get_confidence(statement)

        if confidence > self.confidence_threshold:
            self.load_state()
            active_monitors = self.state.get("active_monitors")
            selected_asset = self.get_selected_asset()

            if not selected_asset:
                response = Statement(text="No asset selected. Please select an asset first.")
                response.confidence = confidence
            else:
                response = CallbackAction(
                    self.execute_action,
                    selected_asset=selected_asset,
                    active_monitors=active_monitors,
                    confidence=confidence,
                )
        else:
            response = None

        return response

    def execute_action(self, selected_asset, active_monitors):
        active_monitors = dict((item.name, item) for item in active_children() if item.name)
        active_monitors = self._start_monitors(selected_asset, active_monitors)

        if len(active_monitors) == 0:
            response_text = f'{selected_asset["asset_name"]} has no registered monitors'
        else:
            monitor_names = list(active_monitors.keys())
            response_text = "{} monitors running ({})".format(
                len(monitor_names), ", ".join(monitor_names)
            )

            self.state = {"active_monitors": monitor_names}
            self.share_state()

        return response_text

    def _start_monitors(self, selected_asset, active_monitors):
        asset_name = self.get_asset_name(selected_asset)
        asset_monitors = self.all_monitors.get(asset_name, {})

        monitors_to_start = dict(
            (name, settings)
            for (name, settings) in asset_monitors.items()
            if not ((name in active_monitors) and (active_monitors[name].is_alive()))
        )

        for name, settings in monitors_to_start.items():
            monitor_settings = settings.copy()
            is_enabled = monitor_settings.get("enabled", False)
            if not is_enabled:
                logging.info(f"Ignoring disabled process '{name}'")
                continue

            process_type = monitor_settings.get("type")
            if process_type not in self.process_handlers:
                logging.error(f"Ignoring unknown process type '{process_type}'")
                continue

            monitor_settings["event_type"] = self.get_event_type(selected_asset)
            monitor_settings["live"] = self.settings["live"]
            monitor_settings["output"]["room"] = {"id": self.room_id}

            logging.debug(f"Starting {name}")
            try:
                process_func = self.process_handlers.get(process_type)
                process_func = agent_function(process_func, name=name, with_state=True)
                process = process_func(monitor_settings, name=name)
                active_monitors[name] = process
                process.start()

            except Exception as e:
                logging.warn(f"Error starting {name}: {e}")

        return active_monitors



# -*- coding: utf-8 -*-
from multiprocessing import Queue, active_children
from functools import partial

from eliot import start_action
from setproctitle import setproctitle
from chatterbot.trainers import ChatterBotCorpusTrainer

from live_client import query
from live_client.events import messenger
from live_client.facades import LiveClient
from live_client.types.message import Message
from live_client.utils import logging

from live_agent.services.processes import agent_function

from live_agent.modules.chatbot.src.bot import ChatBot
from live_agent.modules.chatbot.src.actions import ActionStatement


__all__ = ["start"]

read_timeout = 120
request_timeout = (3.05, 5)


##
# Misc functions
def load_state(container, state_key=None, default=None):
    state = container.get("state", {})

    if state_key and (state_key not in state):
        if default is None:
            default = {}

        share_state(container, state_key=state_key, state_data=default)

    return state


def share_state(container, state_key=None, state_data=None):
    if "state" not in container:
        container.update(state={})

    container["state"].update(**{state_key: state_data})


##
# Chat message handling
def maybe_extract_messages(event):
    event_content = event.get("data", {}).get("content", [])

    return [
        Message(item)
        for item in event_content
        if (item.get("__type") == "__message") and (item.get("message") is not None)
    ]


def maybe_mention(settings, message):
    bot_alias = settings.get("alias", "Intelie")
    is_mention = message.has_mention(bot_alias)
    if is_mention:
        message = message.remove_mentions(bot_alias)

    return is_mention, message


def process_messages(chatbot, messages):
    settings = chatbot.context.get("settings")
    room_id = chatbot.context.get("room_id")

    for message in messages:
        with start_action(action_type="process_message", message=message.get("text")):
            is_mention, message = maybe_mention(settings, message)

            response = None
            if is_mention:
                response = chatbot.get_response(message)

            if response is not None:
                logging.info('Bot response is "{}"'.format(response.serialize()))
                if isinstance(response, ActionStatement):
                    response.chatbot = chatbot
                    response_message = response.run()
                else:
                    response_message = response.text

                maybe_send_message(settings, room_id, response_message)


def maybe_send_message(settings, room_id, response_message):
    bot_settings = settings.copy()
    bot_alias = bot_settings.get("alias", "Intelie")
    bot_settings["output"]["room"] = {"id": room_id}
    if "name" not in bot_settings["output"]["author"]:
        bot_settings["output"]["author"]["name"] = bot_alias

    messenger.send_message(
        response_message, settings=bot_settings, message_type=messenger.MESSAGE_TYPES.CHAT
    )


##
# Room Bot initialization
def train_bot(chatbot, language="english"):
    trainer = ChatterBotCorpusTrainer(chatbot)
    trainer.train(f"chatterbot.corpus.{language}.conversations")
    trainer.train(f"chatterbot.corpus.{language}.greetings")
    trainer.train(f"chatterbot.corpus.{language}.humor")


def start_chatbot(settings, room_id, room_queue, **kwargs):
    setproctitle("DDA: Chatbot for room {}".format(room_id))

    # Load the previous state
    state_manager = kwargs.get("state_manager")
    state = state_manager.load()

    settings.update(state=state.get("bot_state", {}))
    load_state_func = partial(load_state, settings)
    share_state_func = partial(share_state, settings)

    bot_alias = settings.get("alias", "Intelie")
    context = {
        "room_id": room_id,
        "settings": settings,
        "live_client": LiveClient(settings, room_id),
        "functions": {"load_state": load_state_func, "share_state": share_state_func},
    }
    chatbot = ChatBot(
        bot_alias,
        read_only=True,
        prefer_agreement=False,
        logic_adapters=settings.get("logic_adapters", []),
        preprocessors=["chatterbot.preprocessors.clean_whitespace"],
        filters=[],
        **context,
    )
    train_bot(chatbot)

    while True:
        event = room_queue.get()
        messages = maybe_extract_messages(event)
        process_messages(chatbot, messages)
        state_manager.save({"bot_state": settings.get("state", {})}, force=True)

    return chatbot


def add_bot(settings, bots_registry, room_id):
    new_bot = False
    if room_id is not None:
        room_bot, room_queue = bots_registry.get(room_id, (None, None))

        if room_bot and room_bot.is_alive():
            logging.debug("Bot for {} is already known".format(room_id))
        else:
            logging.info("New bot for room {}".format(room_id))
            new_bot = True

            start_chatbot_with_log = agent_function(
                start_chatbot, name=f"bot for room {room_id}", with_state=True
            )
            with start_action(action_type="start_chatbot", room_id=room_id) as action:
                task_id = action.serialize_task_id()
                room_queue = Queue()
                room_bot = start_chatbot_with_log(settings, room_id, room_queue, task_id=task_id)

            room_bot.start()
            bots_registry[room_id] = (room_bot, room_queue)

    return bots_registry, new_bot


def route_message(settings, bots_registry, event):
    logging.debug("Got an event: {}".format(event))

    messages = maybe_extract_messages(event)
    for message in messages:
        room_id = message.get("room", {}).get("id")
        sender = message.get("author", {})

        bots_registry, new_bot = add_bot(settings, bots_registry, room_id)
        if new_bot:
            messenger.add_to_room(settings, room_id, sender)

        # Send the message to the room's bot process
        room_bot, room_queue = bots_registry.get(room_id, (None, None))
        room_queue.put(event)

    return [item[0] for item in bots_registry.values()]


##
# Global process initialization
def start(settings, **kwargs):
    setproctitle("DDA: Chatbot main process")
    logging.info("Chatbot process started")

    # Load the previous state
    state_manager = kwargs.get("state_manager")
    state = state_manager.load()
    bots_registry = state.get("bots_registry", {})

    # Restart previously known bots
    rooms_with_bots = bots_registry.keys()
    for room_id in rooms_with_bots:
        bots_registry, new_bot = add_bot(settings, bots_registry, room_id)

    bot_alias = settings.get("alias", "Intelie").lower()
    bot_query = f"""
        __message -__delete:*
        => @filter(
            message:lower():contains("{bot_alias}") &&
            author->name:lower() != "{bot_alias}"
        )
    """

    @query.on_event(bot_query, settings, timeout=read_timeout)
    def handle_events(event, *args, **kwargs):
        messenger.join_messenger(settings)
        route_message(settings, bots_registry, event)

        # There is no use saving the processes, so we save a dict with no values
        state_manager.save(
            {"bots_registry": dict((room_id, (None, None)) for room_id in bots_registry)},
            force=True,
        )
        return

    try:
        handle_events()
    except Exception:
        # Nothing to do. Let this process end
        pass

    for bot in active_children():
        bot.terminate()
        bot.join(5)

    return



import re

from chatterbot.logic import LogicAdapter

from ..src.actions import CallbackAction, ShowTextAction


class BotVariablesAdapter(LogicAdapter):

    set_command_regex = r"set\s+(.+?)\s*=\s*(.+?)\s*$"
    get_command_regex = r"get\s+(.+?)\s*$"

    def can_process(self, statement):
        words = re.split(r"\s+", statement.text)
        return ("set" in words) or ("get" in words)

    def process(self, statement, additional_response_selection_parameters=None):
        text = statement.text
        m = re.match(self.set_command_regex, text)
        if m is not None:
            name = m.group(1)
            value = m.group(2)
            return CallbackAction(self.handle_set_variable, confidence=1, name=name, value=value)

        m = re.match(self.get_command_regex, text)
        if m is not None:
            name = m.group(1)
            return CallbackAction(self.handle_get_variable, confidence=1, name=name)

        return ShowTextAction(f"Unable to parse command: {text}", confidence=0.5)

    def handle_set_variable(self, name, value):
        self.chatbot.setvar(name, value)
        return f"`> {name} = {value}`"

    def handle_get_variable(self, name):
        value = self.chatbot.getvar(name)
        value = str(value or "Not defined")
        return f"`> {name} = {value}`"



# -*- coding: utf-8 -*-
import csv
from setproctitle import setproctitle

import lasio

from live_client.utils import logging

__all__ = ["start"]


def open_files(settings, iterations):
    path_list = settings["path_list"]
    index_mnemonic = settings["index_mnemonic"]

    path_index = iterations

    try:
        las_path, chat_path = path_list[path_index]
        with open(las_path, "r") as las_file:
            data = lasio.read(las_file)

        if chat_path:
            with open(chat_path, "r") as chat_file:
                chat_data = list(csv.DictReader(chat_file))

            logging.debug("Success opening files {} and {}>".format(las_path, chat_path))
        else:
            chat_data = []
            logging.debug("Success opening file {}>".format(las_path))

        success = True
    except Exception as e:
        data = e
        chat_data = None
        success = False
        logging.error("Error opening file {}, {}<{}>".format(las_path, e, type(e)))

    return success, data, chat_data, index_mnemonic


def export_curves_data(event_type, las_data, chat_data, index_mnemonic, settings):
    logging.info("Exporting curves for {}".format(event_type))
    output_dir = settings.get("temp_dir", "/tmp")

    source_name = las_data.version.SOURCE.value
    output_filename = "{}/{}.csv".format(output_dir, source_name)

    with open(output_filename, "w") as output_file:
        writer = csv.writer(output_file)

        for curve in las_data.curves:
            writer.writerow(
                ["{} - {}".format(curve.mnemonic, curve.descr), curve.mnemonic, curve.unit, "", ""]
            )

    logging.info("File {} created".format(output_filename))


def start(settings, **kwargs):
    event_type = settings["output"]["event_type"]
    setproctitle('DDA: LAS replayer for "{}"'.format(event_type))

    handling_func = export_curves_data

    iterations = 0
    while True:
        try:
            success, las_data, chat_data, index_mnemonic = open_files(settings, iterations)

            if success:
                handling_func(event_type, las_data, chat_data, index_mnemonic, settings)
                logging.info("{}: Iteration {} successful".format(event_type, iterations))

            else:
                logging.info("{}: Single pass mode, exiting".format(event_type))
                break

        except Exception as e:
            logging.error("{}: Error processing events, {}<{}>".format(event_type, e, type(e)))

        iterations += 1

    return



# -*- coding: utf-8 -*-
from pprint import pformat
from chatterbot.logic import LogicAdapter
from chatterbot.conversation import Statement
from eliot import start_action
import nltk
from nltk import NaiveBayesClassifier

from live_client.assets.curves import only_enabled_curves
from live_client.utils import logging
from live_client.query import on_event


__all__ = []


class BaseBayesAdapter(LogicAdapter):
    """
    Superclass for adapters using naive bayes
    """

    positive_examples = []
    confidence_threshold = 0.75
    confidence_damping_thresold = 0.9
    _classifier = None

    @property
    def negative_examples(self):
        examples = []
        my_name = str(self.__class__)
        for adapter in self.chatbot.logic_adapters:
            adapter_name = str(adapter.__class__)
            if (adapter_name != my_name) and hasattr(adapter, "positive_examples"):
                examples.extend(adapter.positive_examples)

        return examples

    @property
    def classifier(self):
        if self._classifier is None:
            self._classifier = self.prepare_classifier()

        return self._classifier

    def prepare_classifier(self):
        labeled_data = []
        labeled_data.extend([(name, 0) for name in self.negative_examples])
        labeled_data.extend([(name, 1) for name in self.positive_examples])

        train_set = [(self.analyze_features(text), n) for (text, n) in labeled_data]
        return NaiveBayesClassifier.train(train_set)

    def prepare_features(self):
        # Extract meaningful features from the examples
        # A list of all words from the known sentences
        all_examples = self.positive_examples + self.negative_examples
        self.all_words = " ".join(all_examples).split()

        # A list of the first word in each of the known sentence
        self.all_first_words = [sentence.split(" ", 1)[0] for sentence in all_examples]

    def analyze_features(self, text):
        """
        Provide an analysis of significant features in the string.
        """
        self.prepare_features()
        features = {}

        for word in text.split():
            features["first_word({})".format(word)] = word in self.all_first_words

        for word in text.split():
            features["contains({})".format(word)] = word in self.all_words

        for letter in "abcdefghijklmnopqrstuvwxyz":
            features["count({})".format(letter)] = text.lower().count(letter)
            features["has({})".format(letter)] = letter in text.lower()

        return features

    def get_confidence(self, statement):
        my_features = self.analyze_features(statement.text.lower())
        return self.classifier.classify(my_features) * self.confidence_damping_thresold

    def process(self, statement, additional_response_selection_parameters=None):
        confidence = self.get_confidence(statement)

        response = Statement(
            text="{}: {}, confidence={}".format(
                self.__class__.__name__, statement.search_text, confidence
            )
        )

        self.confidence = response.confidence = confidence
        return response

    def can_process(self, statement):
        confidence = self.get_confidence(statement)
        can_process = confidence > self.confidence_threshold

        logging.debug(
            """{} (confidence {}):
            The 10 most informative features are: {}
            """.format(
                self.__class__.__name__,
                confidence,
                ", ".join(
                    pformat(item) for item in self.classifier.most_informative_features(n=10)
                ),
            )
        )

        return can_process


class NLPAdapter(LogicAdapter):
    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)
        self.download_and_log("maxent_ne_chunker")
        self.download_and_log("words")
        self.download_and_log("punkt")

    def download_and_log(self, name):
        logging.info(f"Downloading {name}")
        nltk.download(name)

    def tokenize(self, statement):
        return nltk.word_tokenize(statement.text)

    def pos_tag(self, statement):
        tokens = self.tokenize(statement)
        return nltk.pos_tag(tokens)


class WithStateAdapter(LogicAdapter):
    """
    Superclass for adapters requiring an internal state per conversation
    """

    state_key = None
    required_state = []
    default_state = {}

    def __init__(self, chatbot, **kwargs):
        functions = kwargs.get("functions", {})
        if ("load_state" not in functions) or ("share_state" not in functions):
            raise ValueError("Cannot keep state without a reference to state management functions")

        self.__state = None
        self.__shared_state = None
        self._load_state = functions["load_state"]
        self._share_state = functions["share_state"]

        super().__init__(chatbot, **kwargs)

    def load_state(self):
        self.__shared_state = self._load_state(state_key=self.state_key, default=self.default_state)
        self.__state = self.__shared_state[self.state_key]

    def share_state(self):
        self._share_state(state_key=self.state_key, state_data=self.state)

    @property
    def shared_state(self):
        if not self.__shared_state:
            self.load_state()

        return self.__shared_state

    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, new_state):
        self.__state = new_state


class WithAssetAdapter(WithStateAdapter):

    query_timeout = 60

    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)
        self.settings = kwargs.get("settings", {})

    def get_selected_asset(self):
        return self.shared_state.get("selected-asset", {})

    def get_asset_name(self, asset):
        return asset.get("asset_name")

    def get_event_type(self, asset):
        return asset.get("asset_config", {}).get("event_type")

    def get_asset_curves(self, asset):
        all_curves = asset.get("asset_config", {}).get("curves", {})
        return only_enabled_curves(all_curves)

    def curve_was_mentioned(self, curve, statement, exact=True, match_case=True):
        statement_text = statement.text

        if match_case is False:
            statement_text = statement_text.upper()
            curve = curve.upper()

        text_words = statement_text.split()
        if exact:
            result = any(filter(lambda word: curve == word, text_words))
        else:
            result = curve in statement_text

        return result

    def list_mentioned_curves(self, statement):
        selected_asset = self.get_selected_asset()
        curves = self.get_asset_curves(selected_asset)

        # First, try a more lenient search
        mentions = dict(
            (item, {"exact": False})
            for item in curves
            if self.curve_was_mentioned(item, statement, exact=False, match_case=False)
        )

        # Then try to find an exact mention of the curves
        mentions.update(
            dict(
                (item, {"exact": True})
                for item in curves
                if self.curve_was_mentioned(item, statement, exact=True)
            )
        )
        return mentions

    def find_selected_curves(self, statement):
        index_curve = getattr(self, "index_curve", None)

        if index_curve is None:
            mentioned_curves = self.list_mentioned_curves(statement)
        else:
            mentioned_curves = dict(
                (name, data)
                for name, data in self.list_mentioned_curves(statement).items()
                if name != index_curve
            )

        # Try to find an exact mention to a curve
        selected_curves = [
            name for name, match_data in mentioned_curves.items() if match_data.get("exact") is True
        ]

        # Failing that, use all matches
        if not selected_curves:
            selected_curves = list(mentioned_curves.keys())

        return selected_curves

    def run_query(self, query_str, realtime=False, span=None, callback=None):
        settings = self.settings
        timeout = self.query_timeout

        with start_action(action_type=self.state_key, query=query_str):

            @on_event(query_str, settings, span=span, realtime=realtime, timeout=timeout)
            def handle_events(event, callback, *args, **kwargs):
                event_data = event.get("data", {})
                result = event_data.get("content", [])
                return callback(result)

            return handle_events(callback)



# -*- coding: utf-8 -*-

__all__ = ["Monitor"]


class Monitor:
    """Base class to implement monitors"""

    monitor_name = "base_monitor"

    def __init__(self, settings, **kwargs):
        self.settings = settings
        self.kwargs = kwargs

    def run(self):
        raise NotImplementedError("Monitors must define a start method")

    @classmethod
    def start(cls, settings, **kwargs):
        monitor = cls(settings, **kwargs)
        monitor.run()



# -*- coding: utf-8 -*-
from enum import Enum
import csv
from setproctitle import setproctitle

import lasio

from live_client.events import raw, messenger
from live_client.utils import timestamp, logging

from ..utils import loop

__all__ = ["start"]


READ_MODES = Enum("READ_MODES", "SINGLE_PASS, CONTINUOUS")


def update_chat(chat, last_ts, next_ts, index_mnemonic, settings):
    if not chat:
        return

    items_to_send = []
    for item in chat:
        item_index = int(item.get(index_mnemonic, -1))

        if last_ts <= item_index < next_ts:
            items_to_send.append(item)
        elif item_index > next_ts:
            break

    logging.debug("{} messages between {} and {}".format(len(items_to_send), last_ts, next_ts))

    for item in items_to_send:
        message = item.get("MESSAGE", "")
        source = item.get("SOURCE", "")
        if message and source:
            messenger.maybe_send_chat_message(
                message, timestamp.get_timestamp(), settings, author_name=source
            )


def send_message(message, timestamp, settings=None):
    messenger.maybe_send_message_event(message, timestamp, settings)
    messenger.maybe_send_chat_message(message, timestamp, settings)


def delay_output(last_timestamp, next_timestamp):
    if last_timestamp == 0:
        sleep_time = 0
    else:
        sleep_time = max(next_timestamp - last_timestamp, 0)

    loop.await_next_cycle(sleep_time)


def read_next_frame(values_iterator, curves, curves_data, index_mnemonic):
    try:
        index, values = next(values_iterator)
        success = True
    except Exception as e:
        output_frame = {}
        success = False
        logging.debug("Error reading next value, {}<{}>".format(e, type(e)))

    if success:
        output_frame = {index_mnemonic: {"value": index, "uom": "s"}}

        for index, channel in enumerate(curves):
            uom = curves_data.get(channel)
            channel_value = values[index]
            output_frame[channel] = {"value": channel_value, "uom": uom}

    return success, output_frame


def open_files(settings, iterations, mode=READ_MODES.CONTINUOUS):
    path_list = settings["path_list"]
    index_mnemonic = settings["index_mnemonic"]

    if mode == READ_MODES.CONTINUOUS:
        path_index = iterations % len(path_list)
    else:
        path_index = iterations

    try:
        las_path, chat_path = path_list[path_index]
        with open(las_path, "r") as las_file:
            data = lasio.read(las_file)

        if chat_path:
            with open(chat_path, "r") as chat_file:
                chat_data = list(csv.DictReader(chat_file))

            logging.debug("Success opening files {} and {}>".format(las_path, chat_path))
        else:
            chat_data = []
            logging.debug("Success opening file {}>".format(las_path))

        success = True
    except Exception as e:
        data = e
        chat_data = None
        success = False
        logging.error("Error opening file {}, {}<{}>".format(las_path, e, type(e)))

    return success, data, chat_data, index_mnemonic


def generate_events(event_type, las_data, chat_data, index_mnemonic, settings, state_manager):
    logging.info("{}: Event generation started".format(event_type))

    source_name = las_data.version.SOURCE.value
    curves_data = dict((item.mnemonic, item.unit) for item in las_data.curves)
    las_df = las_data.df()
    values_iterator = las_df.iterrows()
    curves = las_df.columns

    success = True
    state = state_manager.load()
    last_timestamp = state.get("last_timestamp", 0)
    if last_timestamp > 0:
        logging.info(f"Skipping to index {last_timestamp}")

    while success:
        success, statuses = read_next_frame(values_iterator, curves, curves_data, index_mnemonic)

        if success:
            next_timestamp = statuses.get(index_mnemonic, {}).get("value", 0)

        if next_timestamp > last_timestamp:
            delay_output(last_timestamp, next_timestamp)

            if last_timestamp == 0:
                message = "Replay from '{}' started at TIME {}".format(source_name, next_timestamp)
                send_message(message, timestamp.get_timestamp(), settings=settings)

            raw.create(event_type, statuses, settings)

            update_chat(chat_data, last_timestamp, next_timestamp, index_mnemonic, settings)
            last_timestamp = next_timestamp
            state_manager.save({"last_timestamp": last_timestamp})


def start(settings, **kwargs):
    """
    Starts the LAS replayer, based on the process settings.

    Replays sensors and chat data.

    :param settings: Parameters for this `las_replayer` instance
    :type settings: dict

    :param state_manager: StateManager injected by `live-agent`
    :type state_manager: live_agent.services.StateManager

    :rtype: NoneType

    The settings for this process have the following format::

      {
        "type": "las_replay",
        "enabled": true,  # Self explanatory
        "index_mnemonic": "TIME",  # Curve used as index for the LAS data
        "path_list": [
          # A list of filename pairs containing the data to be replayed
          [<path for a LAS file>, <path for a CSV file containing the chat logs>],
          ...
        ]
        "output": {
          "event_type": "raw_wellX", The name of the event type which should be sent to Intelie Live
          "author": {
            "id": <user id>  # User id of the author for the messages
            "name": "Linguistics monitor"  # User name of the author for the messages
          },
          "room": {
            "id": <room id>  # Id of the room where the messages should be sent
          },
          "message_event": {
            # Information for generating markers on charts
            "event_type": "raw_wellX",  # Usually the raw event type of the asset being monitored
            "mnemonic": "MSG"  # Mnemonic used for messages normalization, usually named `MSG`
          }
        }
      }

    The LAS file *must* be SPACE delimited.

    The CSV file must contain at least 3 columns:

    - `MESSAGE`: The text of the message
    - `SOURCE`: The name of the message sender
    - A column with the same name a the `index_mnemonic` defined on the process settings,
    used for correlating messages with the events from the LAS file.

    """
    event_type = settings["output"]["event_type"]
    cooldown_time = settings.get("cooldown_time", 300)
    setproctitle('DDA: LAS replayer for "{}"'.format(event_type))

    state_manager = kwargs.get("state_manager")
    state = state_manager.load()
    iterations = state.get("iterations", 0)

    while True:
        try:
            success, las_data, chat_data, index_mnemonic = open_files(
                settings, iterations, mode=READ_MODES.CONTINUOUS
            )

            if success:
                generate_events(
                    event_type, las_data, chat_data, index_mnemonic, settings, state_manager
                )
                logging.info("Iteration {} successful".format(iterations))
            else:
                logging.warn("Could not open files")

            state_manager.save({"last_timestamp": 0, "iterations": iterations}, force=True)
            loop.await_next_cycle(
                cooldown_time,
                message="Sleeping for {:.1f} minutes between runs".format(cooldown_time / 60.0),
                log_func=logging.info,
            )

        except KeyboardInterrupt:
            logging.info("Stopping after {} iterations".format(iterations))
            raise

        except Exception as e:
            logging.error(
                "Error processing events during iteration {}, {}<{}>".format(iterations, e, type(e))
            )

        iterations += 1
        del las_data
        del chat_data

    return



# -*- coding: utf-8 -*-
import os
import signal
import json

from eliot import start_action, to_file, Action
from setproctitle import setproctitle

from live_client.utils import logging
from .services import processes, daemon

__all__ = ["LiveAgent"]

PIDFILE_ENVVAR = "DDA_PID_FILE"
DEFAULT_PIDFILE = "/var/run/live-agent.pid"

LOGFILE_ENVVAR = "DDA_LOG_FILE"
DEFAULT_LOG = "/var/log/live-agent.log"


class LiveAgent(daemon.Daemon):
    def __init__(self, pidfile, settings_file):
        setproctitle("DDA:  Main process")
        self.configure_log()
        self.settings_file = settings_file

        with start_action(action_type="init_daemon") as action:
            task_id = action.serialize_task_id()
            error_logfile = f"{self.logfile}.error"
            return super().__init__(
                pidfile, stdout=self.logfile, stderr=error_logfile, task_id=task_id
            )

    def run(self):
        with Action.continue_task(task_id=self.task_id):
            try:
                with open(self.settings_file, "r") as fd:
                    global_settings = json.load(fd)

                logging_settings = global_settings.get("logging")
                live_settings = global_settings.get("live")

                logging.setup_python_logging(logging_settings)
                logging.setup_live_logging(logging_settings, live_settings)

                agent_processes = processes.start(global_settings)
            except KeyboardInterrupt:
                logging.info("Execution interrupted")
                raise
            except Exception:
                logging.exception("Error processing inputs")
                raise

        for item in agent_processes:
            item.terminate()
            item.join()

    @property
    def logfile(self):
        return os.environ.get(LOGFILE_ENVVAR, DEFAULT_LOG)

    def configure_log(self):
        to_file(open(self.logfile, "ab"))


def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)



# -*- coding: utf-8 -*-
import importlib
from pprint import pformat

from chatterbot.conversation import Statement
from chatterbot.utils import validate_adapter_class, initialize_class
from chatterbot.logic import LogicAdapter
from chatterbot.adapters import Adapter

from jinja2 import Template

from ..src.actions import CallbackAction
from ..constants import FEATURES_DESCRIPTION_TEMPLATE
from .base import WithStateAdapter, BaseBayesAdapter

__all__ = ["BotFeaturesAdapter", "StateDebugAdapter", "AdapterReloaderAdapter"]


class BotFeaturesAdapter(BaseBayesAdapter):
    """
    Lists the capabilities of the bot
    """

    state_key = "bot-features"
    positive_examples = [
        "/help",
        "!help",
        "help me",
        "some help",
        "/commands",
        "!commands",
        "command list",
        "list commands",
        "list your features",
        "which features have",
        "which tasks can perform",
        "how can you help me",
        "how can you help us",
        "what can you do",
        "what can be done you",
        "what is your job",
        "how can you help me",
        "what are you capable",
        "which capabilities do you have",
        "which powers do you have",
        "what is your superpower",
        "which are your superpowers",
        "which superpowers do you have",
    ]

    def prepare_response(self):
        template = Template(FEATURES_DESCRIPTION_TEMPLATE)
        adapters = self.chatbot.logic_adapters
        response_text = template.render(
            bot_name=self.chatbot.name,
            features=[
                item
                for item in adapters
                if hasattr(item, "description") and hasattr(item, "usage_example")
            ],
        )
        return response_text

    def process(self, statement, additional_response_selection_parameters=None):
        self.confidence = self.get_confidence(statement)
        response = Statement(text=self.prepare_response())
        response.confidence = self.confidence

        return response


class StateDebugAdapter(WithStateAdapter):
    """
    Displays the shared state
    """

    keyphrase = "show me your inner self"
    state_key = "state-debug"

    def process(self, statement, additional_response_selection_parameters=None):
        response = Statement(text=pformat(self.shared_state, depth=3))
        response.confidence = 1

        return response

    def can_process(self, statement):
        return self.keyphrase in statement.text.lower()


class AdapterReloaderAdapter(WithStateAdapter):
    """
    Reloads the code for all the logic adapters
    """

    keyphrase = "reinvent yourself"
    state_key = "adapter-reloader"

    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)
        settings = kwargs.get("settings", {})
        self.logic_adapters = settings.get("logic_adapters", [])

    def process(self, statement, additional_response_selection_parameters=None):
        return CallbackAction(self.execute_action, confidence=1)

    def can_process(self, statement):
        return self.keyphrase in statement.text.lower()

    def execute_action(self):
        try:
            self._delete_all_adapters()

            # Reload the list of logic adapters
            self.chatbot.logic_adapters = [
                self._initialize_class(adapter)
                for adapter in self.logic_adapters
                if self._validate_adapter(adapter)
            ]
            response_text = "{} logic adapters reloaded".format(len(self.logic_adapters))

        except Exception as e:
            response_text = "Error reloading adapters: {} {}".format(e, type(e))

        return response_text

    def _delete_all_adapters(self):
        for adapter_instance in self.chatbot.logic_adapters:
            del adapter_instance

    def _initialize_class(self, adapter):
        if isinstance(adapter, dict):
            adapter.pop("logic_adapters", None)
            adapter_path = adapter.get("import_path")
        else:
            adapter_path = adapter

        # Reload the logic adapters
        module_parts = adapter_path.split(".")
        module_path = ".".join(module_parts[:-1])
        module = importlib.import_module(module_path)
        module = importlib.reload(module)

        return initialize_class(adapter, self.chatbot, **self.chatbot.context)

    def _validate_adapter(self, adapter):
        try:
            validate_adapter_class(adapter, LogicAdapter)
            is_valid = True
        except Adapter.InvalidAdapterTypeException:
            is_valid = False

        return is_valid



# -*- coding: utf-8 -*-
from functools import partial

from chatterbot.conversation import Statement  # NOQA
from eliot import start_action

from live_agent.modules.chatbot.src.actions import CallbackAction
from live_agent.modules.chatbot.logic_adapters.base import (
    BaseBayesAdapter,
    NLPAdapter,
    WithAssetAdapter,
)

ITEM_PREFIX = "\n"


__all__ = ["TradesCounterAdapter"]


class TradesCounterAdapter(BaseBayesAdapter, NLPAdapter, WithAssetAdapter):
    """
    Counts the number of trades over the last X minutes
    """

    state_key = "trades-count-query"
    default_state = {}
    positive_examples = ["number of trades over", "count the trades on"]
    description = "Counts the number of trades over a number of minutes"
    usage_example = "count the number of trades over the last [number] minutes"

    def format_response(self, response_content, interval=None):
        if not response_content:
            result = f"No trades recorded over the last {interval} minutes"
        else:
            results = [f"Trades over the last {interval} minutes:"]
            for item in response_content:
                results.append(f"- *{item['pair']}*: {item['count']:.0f} trades")

            result = ITEM_PREFIX.join(results)

        return result

    def run_query(self, interval):
        value_query = "krakenfx => count() by pair at the end => @filter(count != null)"

        return super().run_query(
            value_query,
            realtime=False,
            span=f"last {interval} minutes",
            callback=partial(self.format_response, interval=interval),
        )

    def find_interval_value(self, statement):
        tagged_words = self.pos_tag(statement)

        # look for a number in the statement
        value = None
        for word, tag in tagged_words:
            if tag == "CD":  # CD: Cardinal number
                value = word
                break

        return value

    def can_process(self, statement):
        can_process = super().can_process(statement)
        words = statement.text.lower().split(" ")
        return can_process and ("minutes" in words)

    def process(self, statement, additional_response_selection_parameters=None):
        confidence = self.get_confidence(statement)
        if confidence > self.confidence_threshold:
            return CallbackAction(self.execute_action, confidence, statement=statement)

    def execute_action(self, statement):
        interval = self.find_interval_value(statement)

        if (interval is not None) and (int(interval) > 0):
            with start_action(action_type=self.state_key, interval=interval):
                response_text = self.run_query(interval)

        else:
            response_text = "I'm sorry, I didn't get the interval you want me to count"

        return response_text



import os
from setuptools import setup, find_packages

__all__ = ["setup"]


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ""


def get_readme():
    """Return the README file contents. Supports text,rst, and markdown"""
    for name in ("README", "README.rst", "README.md"):
        if os.path.exists(name):
            return read_file(name)
    return ""


def install_requires(module=None):
    if module:
        filepath = f"./live_agent/modules/{module}/"
    else:
        filepath = ""

    requirements = read_file(f"{filepath}requirements.txt")
    return requirements


setup(
    name="live_agent",
    version="0.11.1",
    description="A framework for implementing agents which interact with the Intelie LIVE platform",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    scripts=[
        "live_agent/scripts/agent-control",
        "live_agent/scripts/create-agent",
        "live_agent/scripts/add-agent-module",
        "live_agent/scripts/validate-settings",
    ],
    url="https://github.com/intelie/live-agent",
    author="Vitor Mazzi",
    author_email="vitor.mazzi@intelie.com.br",
    install_requires=[
        "live-client>=0.8.1",
        "eliot>=1.12.0",
        "eliot-tree>=19.0.0",
        "setproctitle>=1.1.10",
        "dill>=0.3.1.1",
        "requests>=2,<3",
    ],
    extras_require={
        "chatbot": [
            "ChatterBot==1.0.5",
            "chatterbot-corpus==1.2.0",
            "Jinja2==2.11.3",
            "pytz>=2019.2",
            "python-dateutil>=2.7,<2.8",
            "PyYAML>=3.12,<4.0",
        ],
        "las": ["lasio==0.23", "pandas==0.24.2", "scikit-learn>=0.20"],
    },
    zip_safe=False,
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
)



# -*- coding: utf-8 -*-

from live_client.utils import logging

__all__ = ["prepare_query", "handle_events"]


def prepare_query(settings):
    event_type = settings.get("event_type")
    mnemonics_settings = settings.get("monitor", {}).get("mnemonics", {})
    query_mnemonics = list(mnemonics_settings.values())

    mnemonics_list = "|".join(query_mnemonics)
    values_pipe_fragments = [
        r"lastv(value:object):if(\mnemonic:{0}) as {0}".format(item) for item in query_mnemonics
    ]
    units_pipe_fragments = [
        r"lastv(uom:object):if(\mnemonic:{0}) as {0}_uom".format(item) for item in query_mnemonics
    ]
    pipe_fragments = values_pipe_fragments + units_pipe_fragments
    mnemonics_pipe = ", ".join(pipe_fragments)

    query = """
        {} mnemonic!:({}) .flags:nocount
        => {} over last second every second
        => @filter({} != null)
    """.format(
        event_type, mnemonics_list, mnemonics_pipe, query_mnemonics[0]
    )
    logging.debug(f'query is "{query}"')

    return query


def handle_events(event, callback, settings, accumulator=None):
    monitor_settings = settings.get("monitor", {})
    window_duration = monitor_settings.get("window_duration", 60)
    mnemonics = monitor_settings.get("mnemonics", {})
    index_mnemonic = mnemonics.get("index", "timestamp")

    if accumulator is None:
        accumulator = []

    try:
        latest_data, missing_curves = validate_event(event, settings)

        if latest_data:
            accumulator, start, end = refresh_accumulator(
                latest_data, accumulator, index_mnemonic, window_duration
            )

            if accumulator:
                callback(accumulator)

        elif missing_curves:
            missing_curve_names = ", ".join(missing_curves)
            logging.info(f"Some curves are missing ({missing_curve_names}) from event {event} ")

    except Exception as e:
        logging.exception(f"Error during query: <{e}>")
        handle_events(event, callback, settings)
        return


def validate_event(event, settings):
    valid_events = []
    mnemonics_settings = settings.get("monitor", {}).get("mnemonics", {})
    expected_curves = set(mnemonics_settings.values())

    event_content = event.get("data", {}).get("content", [])
    if event_content:
        missing_curves = expected_curves
        for item in event_content:
            item_curves = set(item.keys())

            # Which curves are missing from all items in this event?
            missing_curves = missing_curves - item_curves

            # Does this item has all curves?
            is_valid = len(expected_curves - item_curves) == 0
            if is_valid:
                valid_events.append(item)
    else:
        missing_curves = []

    return valid_events, missing_curves


def refresh_accumulator(latest_events, accumulator, index_mnemonic, window_duration):
    # Purge old events and add the new ones
    latest_event = latest_events[-1]
    window_end = latest_event.get(index_mnemonic, 0)
    window_start = window_end - window_duration
    last_index = window_start

    if index_mnemonic not in latest_event:
        mnemonics_list = latest_event.keys()
        logging.error(
            f"Mnemonic '{index_mnemonic}' not found. Available mnemonics are: '{mnemonics_list}'"
        )

    accumulator.extend(latest_events)
    purged_accumulator = []
    for item in accumulator:
        index = item.get(index_mnemonic, 0)
        if (window_start <= index <= window_end) and (index >= last_index):
            purged_accumulator.append(item)
            last_index = index
        elif index < last_index:
            # Reset the accumulator
            purged_accumulator = [item]
            last_index = index
        elif index == 0:
            logging.error(f"{index_mnemonic} not found, ignoring event")

    logging.debug(
        "{} of {} events between {} and {}".format(
            len(purged_accumulator), len(accumulator), window_start, window_end
        )
    )

    return purged_accumulator, window_start, window_end



# -*- coding: utf-8 -*-

ITEM_PREFIX = "\n  "
SELECTED_ASSET_VARIABLE_NAME = "selected-asset"

FEATURES_DESCRIPTION_TEMPLATE = """My features are:
{% for feature_data in features %}
  *{{ feature_data.description }}*
  Try: "{{ bot_name }}, {{ feature_data.usage_example }}"
{% endfor %}
"""



# -*- coding: utf-8 -*-
from functools import partial
import time
import json
from uuid import uuid4

from chatterbot.conversation import Statement  # NOQA
from eliot import start_action

from live_client.events import annotation
from live_client.assets import list_assets, fetch_asset_settings
from live_client.assets import run_analysis
from live_client.assets.curves import only_enabled_curves
from live_client.utils import logging
from live_client.utils.timestamp import get_timestamp
from live_client.events.constants import UOM_KEY, VALUE_KEY, TIMESTAMP_KEY

from live_agent.modules.chatbot.src.actions import CallbackAction, ShowTextAction
from live_agent.modules.chatbot.logic_adapters.base import (
    BaseBayesAdapter,
    NLPAdapter,
    WithAssetAdapter,
    WithStateAdapter,
)

from ..constants import ITEM_PREFIX, SELECTED_ASSET_VARIABLE_NAME


__all__ = ["AssetListAdapter", "AssetSelectionAdapter", "AutoAnalysisAdapter"]


class AssetListAdapter(BaseBayesAdapter, WithStateAdapter):
    """
    Interacts with the user to associate the chatbot to an asset
    """

    state_key = "asset-list"
    default_state = {}
    positive_examples = [
        "which assets exist",
        "list the assets",
        "list which assets exist",
        "show the assets",
        "show the assets list",
        "show which assets exist",
        "display the assets",
        "display the assets list",
        "display which assets exist",
    ]
    description = "List the assets available"
    usage_example = "show me the list of assets"

    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)

        available_assets = list_assets(kwargs["settings"])

        if not available_assets:
            logging.warn(f"No assets available. Check permissions for this user!")

        self.load_state()
        self.state = {
            "assets": available_assets,
            "asset_names": [item.get("name") for item in available_assets if "name" in item],
        }
        self.share_state()

    def process(self, statement, additional_response_selection_parameters=None):
        self.confidence = self.get_confidence(statement)

        if self.confidence > self.confidence_threshold:
            response = Statement(
                text="The known assets are:{}{}".format(
                    ITEM_PREFIX, ITEM_PREFIX.join(self.state.get("asset_names", []))
                )
            )
            response.confidence = self.confidence

        else:
            response = None

        return response


class AssetSelectionAdapter(BaseBayesAdapter, WithStateAdapter):
    """
    Interacts with the user to associate the chatbot to an asset
    """

    state_key = SELECTED_ASSET_VARIABLE_NAME
    required_state = ["asset_id", "asset_type", "asset_name", "asset_config"]
    default_state = {}
    positive_examples = [
        "activate asset",
        "select asset",
        "set active asset",
        "update asset",
        "new asset is",
        "update asset to",
        "change asset to",
    ]
    description = "Select an asset for this room"
    usage_example = "activate the asset {asset name}"

    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)

        settings = kwargs["settings"]

        self.asset_fetcher = partial(fetch_asset_settings, settings)

    def was_asset_mentioned(self, asset, statement):
        return asset.get("name", "INVALID ASSET NAME").lower() in statement.text.lower()

    def extract_asset_names(self, statement):
        asset_list = self.shared_state.get("asset-list", {})
        assets = list(
            filter(
                lambda asset: self.was_asset_mentioned(asset, statement),
                asset_list.get("assets", [{}]),
            )
        )
        return assets

    def process(self, statement, additional_response_selection_parameters=None):
        self.load_state()
        self.confidence = self.get_confidence(statement)

        response = None
        if self.confidence > self.confidence_threshold:
            selected_assets = self.extract_asset_names(statement)
            num_selected_assets = len(selected_assets)

            if num_selected_assets == 0:
                self.confidence_threshold *= 0.7
                response = ShowTextAction(
                    "I didn't get the asset name. Can you repeat please?", self.confidence
                )

            elif num_selected_assets == 1:
                response = CallbackAction(
                    self.execute_action, self.confidence, asset_info=selected_assets[0]
                )

            elif num_selected_assets > 1:
                response = ShowTextAction(
                    "I didn't understand, which of the assets you meant?{}{}".format(
                        ITEM_PREFIX, ITEM_PREFIX.join(item.get("name") for item in selected_assets)
                    ),
                    self.confidence,
                )
        return response

    def execute_action(self, asset_info):
        asset_name = asset_info.get("name")
        asset_id = asset_info.get("id", 0)
        asset_type = asset_info.get("asset_type", "rig")
        asset_config = self.asset_fetcher(asset_id, asset_type=asset_type)

        if asset_config is None:
            return f"Error fetching information about {asset_name}"

        self.state = {
            "asset_id": asset_id,
            "asset_type": asset_type,
            "asset_name": asset_name,
            "asset_config": asset_config,
        }
        self.share_state()

        event_type = asset_config.get("event_type", None)
        asset_curves = only_enabled_curves(asset_config.get("curves", {}))

        text_templ = (
            "Ok, the asset {} was selected." '\nIt uses the event_type "{}" and has {} curves'
        )
        return text_templ.format(asset_name, event_type, len(asset_curves.keys()))


class AutoAnalysisAdapter(BaseBayesAdapter, NLPAdapter, WithAssetAdapter):
    """
    Analyze a curve on live
    """

    state_key = "auto-analysis"
    required_state = ["assetId", "channel", "begin", "end", "computeFields"]
    default_state = {"computeFields": ["min", "max", "avg", "stdev", "linreg", "derivatives"]}
    positive_examples = [
        "analyse",
        "analyse mnemonic",
        "analyse curve",
        "analyse mnemonic",
        "analyse curve",
        "run an analysis on",
        "execute an analysis on",
        "can you analyse mnemonic",
        "can you analyse curve",
        "can you run an analysis on",
        "can you execute an analysis on",
    ]
    description = "Run an analysis on a curve"
    usage_example = "run an analysis on {curve name}"

    def __init__(self, chatbot, **kwargs):
        super().__init__(chatbot, **kwargs)
        settings = kwargs.get("settings", {})
        self.annotator = partial(annotation.create, settings=settings)

        self.room_id = kwargs["room_id"]
        self.analyzer = partial(run_analysis, settings)

    def run_analysis(self, asset, curve, begin=None, duration=30000):
        if begin is None:
            begin = get_timestamp() - duration

        end = begin + duration

        analysis_results = self.analyzer(
            assetId="{0[asset_type]}/{0[asset_id]}".format(asset),
            channel=curve,
            qualifier=curve,
            computeFields=self.default_state.get("computeFields"),
            begin=begin,
            end=end,
        )

        if analysis_results:
            # Gerar annotation
            analysis_results.update(
                __src="auto-analysis", uid=str(uuid4()), createdAt=get_timestamp()
            )
            with start_action(action_type="create annotation", curve=curve):
                self.annotator(analysis_results, room={"id": self.room_id})

            response_text = "Analysis of curve {} finished".format(curve)

        else:
            response_text = "Analysis of curve {} returned no data".format(curve)

        return response_text

    def process_analysis(self, statement, selected_asset, begin=None):
        selected_curves = self.find_selected_curves(statement)
        num_selected_curves = len(selected_curves)

        if num_selected_curves == 1:
            selected_curve = selected_curves[0]

            ##
            # Iniciar analise
            with start_action(action_type=self.state_key, curve=selected_curve):
                response_text = self.run_analysis(selected_asset, selected_curve)

        elif num_selected_curves == 0:
            response_text = "I didn't get the curve name. Can you repeat please?"

        else:
            response_text = "I'm sorry, which of the curves you want to analyse?{}{}".format(
                ITEM_PREFIX, ITEM_PREFIX.join(selected_curves)
            )

        return response_text

    def process(self, statement, additional_response_selection_parameters=None):
        confidence = self.get_confidence(statement)

        if confidence > self.confidence_threshold:
            self.load_state()
            selected_asset = self.get_selected_asset()
            if selected_asset == {}:
                response_text = "No asset selected. Please select an asset first."
            else:
                return CallbackAction(
                    self.execute_action,
                    confidence,
                    statement=statement,
                    selected_asset=selected_asset,
                )

            response = Statement(text=response_text)
            response.confidence = confidence
        else:
            response = None

        return response

    def execute_action(self, statement, selected_asset):
        return self.process_analysis(statement, selected_asset)


class CurrentValueQueryAdapter(BaseBayesAdapter, NLPAdapter, WithAssetAdapter):
    """
    Returns the current value for a mnemonic
    """

    state_key = "current-query"
    default_state = {}
    positive_examples = ["current value", "value now"]
    description = "Query the most recent value for a curve"
    usage_example = "what is the current value for {curve name}?"

    def format_response(self, response_content, target_curve=None):
        if not response_content:
            result = "No information about {target_curve}".format(target_curve=target_curve)
        else:
            results = []
            for item in response_content:
                query_result = json.loads(item.get(target_curve, "{}"))
                timestamp = int(item.get(TIMESTAMP_KEY, 0)) or None

                try:
                    value = query_result.get(VALUE_KEY)
                    uom = query_result.get(UOM_KEY)

                    if uom:
                        query_result = "{0:.2f} {1}".format(value, uom)
                    else:
                        query_result = "{0:.2f}".format(value)

                except Exception as e:
                    logging.error("{}: {} ({})".format(self.__class__.__name__, e, type(e)))

                if timestamp:
                    time_diff = time.time() - (timestamp / 1000)

                if timestamp < 2:
                    response_age = f"{time_diff:.1f} second ago"
                else:
                    response_age = f"{time_diff:.1f} seconds ago"

                results.append(f"{target_curve} was *{query_result}* {response_age}.")

            result = ITEM_PREFIX.join(results)

        return result

    def run_query(self, target_curve):
        selected_asset = self.get_selected_asset()
        if selected_asset:
            asset_config = selected_asset.get("asset_config", {})

            value_query = """
            {event_type} .flags:nocount .flags:reversed
            => @filter({{{target_curve}}} != null)
            => {{{target_curve}}}:map():json() as {{{target_curve}}}
            """.format(
                event_type=asset_config["filter"], target_curve=target_curve
            )

            return super().run_query(
                value_query,
                realtime=False,
                span="since ts 0 #partial='1'",
                callback=partial(self.format_response, target_curve=target_curve),
            )

    def can_process(self, statement):
        can_process = super().can_process(statement)
        words = statement.text.lower().split(" ")
        return can_process and ("now" in words or "current" in words)

    def process(self, statement, additional_response_selection_parameters=None):
        confidence = self.get_confidence(statement)
        if confidence > self.confidence_threshold:
            self.load_state()
            selected_asset = self.get_selected_asset()

            if selected_asset == {}:
                return ShowTextAction(
                    "No asset selected. Please select an asset first.", confidence
                )
            else:
                return CallbackAction(
                    self.execute_action,
                    confidence,
                    statement=statement,
                    selected_asset=selected_asset,
                )

    def execute_action(self, statement, selected_asset):
        selected_curves = self.find_selected_curves(statement)
        num_selected_curves = len(selected_curves)

        if num_selected_curves == 0:
            response_text = "I didn't get the curve name. Can you repeat please?"

        elif num_selected_curves == 1:
            selected_curve = selected_curves[0]

            with start_action(action_type=self.state_key, curve=selected_curve):
                response_text = self.run_query(selected_curve)

        else:
            response_text = "I'm sorry, which of the curves you meant?{}{}".format(
                ITEM_PREFIX, ITEM_PREFIX.join(selected_curves)
            )

        return response_text



# -*- coding: utf-8 -*-
from typing import Mapping, Iterable, Callable, Optional, Any
from multiprocessing import get_context as get_mp_context
from dataclasses import dataclass
from time import sleep

from eliot import Action, start_action
from live_client.utils import logging

from .importer import load_process_handlers
from .state import StateManager

__all__ = ["start", "agent_function"]


@dataclass
class ProcessSpec:
    function: Callable
    settings: Mapping
    process: Any


def filter_dict(source_dict: Mapping, filter_func: Callable) -> Mapping:
    return dict((key, value) for key, value in source_dict.items() if filter_func(key, value))


def resolve_process_handler(process_type: str, process_handlers: Mapping) -> Mapping:
    return process_handlers.get(process_type)


def get_processes(global_settings: Mapping, process_handlers: Mapping) -> Mapping:
    processes = filter_dict(
        global_settings.get("processes", {}), lambda _k, v: v.get("enabled") is True
    )

    invalid_processes = filter_dict(
        processes, lambda _k, v: (v.get("type") not in process_handlers)
    )

    for name, info in invalid_processes.items():
        logging.error("Invalid process configured: {}, {}".format(name, info))

    valid_processes = filter_dict(processes, lambda name, _v: name not in invalid_processes)

    return valid_processes


def resolve_process_handlers(global_settings: Mapping):
    process_handlers = load_process_handlers(global_settings)
    registered_processes = get_processes(global_settings, process_handlers)

    for name, settings in registered_processes.items():
        process_type = settings.pop("type")
        process_func = resolve_process_handler(process_type, process_handlers)
        settings.update(
            process_func=process_func,
            live=global_settings.get("live", {}),
            process_handlers=process_handlers,
        )

    return registered_processes


def start(global_settings: Mapping) -> Iterable:
    processes_to_run = resolve_process_handlers(global_settings)
    num_processes = len(processes_to_run)
    logging.info(
        "Starting {} processes: {}".format(num_processes, ", ".join(processes_to_run.keys()))
    )

    process_map = {}
    for name, settings in processes_to_run.items():
        process_func = settings.pop("process_func")

        process_map[name] = ProcessSpec(
            function=agent_function(process_func, name=name, with_state=True),
            settings=settings,
            process=None,
        )

    return monitor_processes(process_map)


def monitor_processes(process_map: Mapping, heartbeat_interval: int = 60) -> Iterable:

    while True:
        for name, process_data in process_map.items():
            process = process_data.process

            if process and process.is_alive():
                logging.info(f'Process for "{name}" (pid={process.pid}) is alive')

            else:
                if process:
                    logging.info(f'Process for "{name}" (pid={process.pid}) has died. Restarting')
                else:
                    logging.info(f'Starting "{name}" using {process_data.function}')

                process = process_data.function(process_data.settings)
                try:
                    process.start()
                    logging.info(f'Process for "{name}" (pid={process.pid}) started')
                except OSError as e:
                    logging.exception(f"Error starting process {name} ({e})")

            process_data.process = process

        sleep(heartbeat_interval)

    running_processes = [item["process"] for item in process_map.values()]
    return running_processes


def agent_function(f: Callable, name: Optional[str] = None, with_state: bool = False) -> Callable:
    mp = get_mp_context("fork")

    def wrapped(*args, **kwargs):
        try:
            f_in_action = inside_action(f, name=name, with_state=with_state)
            return mp.Process(target=f_in_action, args=args, kwargs=kwargs)
        except Exception as e:
            logging.exception(f"Error during the execution of {f}: <{e}>")

    return wrapped


def inside_action(f: Callable, name: Optional[str] = None, with_state: bool = False) -> Callable:
    if name is None:
        name = f"{f.__module__}.{f.__name__}"

    def wrapped(*args, **kwargs):
        task_id = kwargs.get("task_id")
        if task_id:
            action = Action.continue_task(task_id=task_id)
        else:
            action = start_action(action_type=name)

        with action.context():
            task_id = action.serialize_task_id()
            kwargs["task_id"] = task_id
            if with_state:
                kwargs["state_manager"] = StateManager(name)

            try:
                return f(*args, **kwargs)
            except Exception as e:
                logging.exception(f"Error during the execution of {f}: <{e}>")

        action.finish()

    return wrapped



import chatterbot

__all__ = ["ChatBot"]


class ChatBot(chatterbot.ChatBot):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.prefer_agreement = kwargs.pop("prefer_agreement", True)
        self.live_client = kwargs.pop("live_client", None)
        self.session = kwargs.pop("session", {})
        self.context = kwargs

    def generate_response(self, input_statement, additional_response_selection_parameters=None):
        """
        Return a response based on a given input statement.

        :param input_statement: The input statement to be processed.
        """
        results = []
        result = None
        max_confidence = -1

        for adapter in self.logic_adapters:
            if adapter.can_process(input_statement):

                output = adapter.process(input_statement, additional_response_selection_parameters)
                results.append(output)

                self.logger.info(
                    '{} selected "{}" as a response with a confidence of {}'.format(
                        adapter.class_name, output.text, output.confidence
                    )
                )

                if output.confidence > max_confidence:
                    result = output
                    max_confidence = output.confidence
            else:
                self.logger.info("Not processing the statement using {}".format(adapter.class_name))

        if self.prefer_agreement:
            result = self.get_most_common_result(results) or result

        # Update the result to return:
        result.in_response_to = input_statement.text
        result.conversation = input_statement.conversation
        result.persona = "bot:" + self.name

        return result

    def get_most_common_result(self, results):
        result = None

        class ResultOption:
            def __init__(self, statement, count=1):
                self.statement = statement
                self.count = count

        # If multiple adapters agree on the same statement,
        # then that statement is more likely to be the correct response
        if len(results) >= 3:
            result_options = {}
            for result_option in results:
                result_string = result_option.text + ":" + (result_option.in_response_to or "")

                if result_string in result_options:
                    result_options[result_string].count += 1
                    statement_confidence = result_options[result_string].statement.confidence
                    if statement_confidence < result_option.confidence:
                        result_options[result_string].statement = result_option
                else:
                    result_options[result_string] = ResultOption(result_option)

            most_common = list(result_options.values())[0]

            for result_option in result_options.values():
                if result_option.count > most_common.count:
                    most_common = result_option

            if most_common.count > 1:
                result = most_common.statement

        return result

    def reset_session(self):
        self.session.clear()

    def setvar(self, name, value):
        self.session[name] = value
        return self

    def getvar(self, name):
        return self.session.get(name)
